package org.netkernel.mod.performance.cache;

import java.util.*;

import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;

public class CachableComputeLightMultiImpulseNormalDataSet extends StandardAccessorImpl
{
	private Random mRandom = new Random(0);
	private List<Impulse> mImpulses = new ArrayList<Impulse>();
	
	/** probability per cycle of creating a new impulse */
	private final float IMPULSE_PROBABILITY=0.01f;
	/** once impulse decays to this level we forget about it */
	private final float IMPULSE_INSIGNIFICANT=0.01f;
	/** impulse decays each step by multiplcation by this value */
	private final float IMPULSE_DECAY=0.99f;
	
	public CachableComputeLightMultiImpulseNormalDataSet()
	{	//declareThreadSafe();
	}
	
	public void onSource(INKFRequestContext aHelper) throws Exception
	{
		INKFRequestReadOnly thisReq=aHelper.getThisRequest();
		String kbString=thisReq.getArgumentValue("Kb");
		String setSizeString=thisReq.getArgumentValue("setSize");
		int setSize=Integer.parseInt(setSizeString);
		
		if (mImpulses.size()==0 || mRandom.nextFloat()<IMPULSE_PROBABILITY)
		{	mImpulses.add(newImpulse(setSize));
		}
		
		float totalAmplitude=0.0f;
		for (Iterator<Impulse> i=mImpulses.iterator(); i.hasNext(); )
		{
			Impulse imp = i.next();
			float amplitude = imp.timestep();
			if (amplitude<IMPULSE_INSIGNIFICANT)
			{	i.remove();
			}
			else
			{	totalAmplitude+=amplitude;
			}
		}
		
		// choose an impulse weighted by amplitude
		float f = mRandom.nextFloat()*totalAmplitude;
		Impulse impulse=null;
		totalAmplitude=0.0f;
		for (Iterator<Impulse> i=mImpulses.iterator(); i.hasNext(); )
		{
			Impulse imp = i.next();
			float amplitude = imp.getAmplitude();
			totalAmplitude+=amplitude;
			if (f<totalAmplitude)
			{	impulse=imp;
				break;
			}
		}
		
		
		if (impulse!=null)
		{
			int item = impulse.getValue();
			
			//System.out.println("impulse "+impulse.getOrigin()+" "+item+"   "+mImpulses.size());
			
			Class repClass=thisReq.getRepresentationClass();
			INKFResponseReadOnly res=aHelper.sourceForResponse("res:/cache/CachableComputeLightItem/"+kbString+"/"+Integer.toString(item),repClass);
			INKFResponse resp=aHelper.createResponseFrom(res.toString());
			resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			//System.out.println((mC++)+" "+item);
		}
		/*
		else
		{	INKFResponse resp=aHelper.createResponseFrom("nothing");
			resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
		}
		*/
	}
	
	private int mC;
	
	private Impulse newImpulse(int aSetSize)
	{
		int origin = (int)(mRandom.nextFloat()*aSetSize);
		int deviation = aSetSize/64;
		float amplitude = 0.25f*0.75f*mRandom.nextFloat();
		return new Impulse(origin,deviation,amplitude);
	}
	
	
	
	private class Impulse
	{
		private final int mOrigin;
		private final float mDeviation;
		private float mAmplitude;
		
		public Impulse(int aOrigin, int aDeviation, float aAmplitude)
		{	mOrigin=aOrigin;
			mDeviation=aDeviation/2;
			mAmplitude=aAmplitude;
		}
		
		public int getOrigin()
		{	return mOrigin;
		}
		
		public float timestep()
		{	mAmplitude*=0.99f;
			return mAmplitude;
		}
		
		public float getAmplitude()
		{	return mAmplitude;
		}
		
		public int getValue()
		{
			float g=(float)mRandom.nextGaussian();
			return mOrigin+(int)(mDeviation*(1.0f+g));
		}
		
	}
}
