package org.netkernel.mod.performance.harness;

import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.ThreadMXBean;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.container.IKernel;
import org.netkernel.container.impl.Kernel;
import org.netkernel.layer0.nkf.INKFAsyncRequestHandle;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.PropertyConfiguration;
import org.netkernel.layer0.util.Utils;

public class RunTestsAccessor extends StandardAccessorImpl
{	
	private boolean mWarmedUp;
	
	public RunTestsAccessor()
	{	//declareThreadSafe(); 
	}	
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		if (!mWarmedUp)
		{	doWarmup(aContext);
		}
		
		String id=aContext.getThisRequest().getArgumentValue("id");
		if (id.equals("sync"))
		{	doSync(aContext);
		}
		else if (id.equals("async"))
		{	doAsync(aContext);
		}
		else if (id.equals("load"))
		{	doLoad(aContext);
		}
		else if (id.equals("cache"))
		{	
			HDSBuilder b = new HDSBuilder();
			b.pushNode("cache");
			doCache(aContext,"CachableComputeLightDataSet",b);
			doCache(aContext,"CachableComputeLightNormalDataSet",b);

			IHDSNode data=b.getRoot();
			aContext.sink("fpds:/tools/nkmark10/test/"+aContext.getThisRequest().getArgumentValue("id"),data);
			INKFResponse resp=aContext.createResponseFrom(data);
			//resp.setExpiry(INKFResponse.EXPIRY_CONSTANT,System.currentTimeMillis()+60000);
		}
		else if (id.equals("gc"))
		{	
			int c=(Integer)aContext.source("res:/mark/gc/5000",Integer.class);
		}
	}
		
	public void doWarmup(INKFRequestContext aContext) throws Exception
	{
		mWarmedUp=true;
		
		//initialise/warmup
		mThreadTimes.clear();
		INKFRequest req=aContext.createRequest("active:countTest");
		req.addArgumentByValue("identifier", "res:/async/test1");
		req.addArgumentByValue("period", 5000L);
		aContext.issueRequest(req);		
	}
	
	private void doCache(INKFRequestContext aContext, String aName, HDSBuilder b) throws Exception
	{
		//blast with CPU*2 threads the computeset and take the best throughput
		
		
		Runtime rt = Runtime.getRuntime();
		int threadCount=rt.availableProcessors()*2;
		//long heapKb=rt.maxMemory()/1024;
		
		MemoryPoolMXBean oldGen = Utils.getOldGen();
		long heapKb=oldGen.getUsage().getMax()/1024;
		//System.out.println("heapKb="+heapKb);
		
		
		INKFAsyncRequestHandle[] handle = new INKFAsyncRequestHandle[threadCount];
		aContext.getKernelContext().getKernel().getRepresentationCache().clear();
		
		int N=16;
		int maxDataset=1024;
		long repKb=2*heapKb/(3*maxDataset);
		b.pushNode(aName);
		for (int c=1; c<N; c++)
		{
			
			int datasetSize=maxDataset*c/N;
		
			//String id="active:relay+id@res:/cache/"+aName+"/"+repKb+"/"+datasetSize;
			String id="res:/cache/"+aName+"/"+repKb+"/"+datasetSize;
			System.out.println(id);
			
			long period=2000L;
			long totalRequest=0;
			for (int n=0; n<2; n++)
			{
				for (int t=0; t<threadCount; t++)
				{	
					INKFRequest req=aContext.createRequest("active:countTest");
					req.addArgumentByValue("identifier", id);
					req.addArgumentByValue("period", period);
					req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
					handle[t]=aContext.issueAsyncRequest(req);
				}
				totalRequest=0;
				for (int t=0; t<threadCount; t++)
				{	long result=(Long)handle[t].join();
					totalRequest+=result;
				}
				
				period/=2;
			}
			
			
			//int cpuUsage=sampleCPU();
			long throughput=totalRequest*1000L/period;
			System.out.println("cache("+c+") "+throughput);
			
			b.pushNode("test");
			b.addNode("count",c);
			b.addNode("throughput",throughput);
			b.popNode();
		}
		
		b.popNode();
		
	}
	
	private void doLoad(INKFRequestContext aContext) throws Exception
	{
		long[] loads=new long[]{1,10000,1000000};
		int originalThreadCount=aContext.getKernelContext().getKernel().getConfiguration().getInt("netkernel.scheduler.threadcount");
		int availableProc=Runtime.getRuntime().availableProcessors();
		setKernelProperty("netkernel.scheduler.threadcount",Integer.toString(availableProc),aContext);	
		int maxThreads=32; 
		long period=4000L;
		HDSBuilder b=new HDSBuilder();
		b.pushNode("load");
		
		for (int i=0; i<loads.length; i++)
		{
		}
		
		b.popNode();
		IHDSNode data=b.getRoot();
		setKernelProperty("netkernel.scheduler.threadcount",Integer.toString(originalThreadCount),aContext);
		aContext.sink("pds:/tools/nkmark10/test/"+aContext.getThisRequest().getArgumentValue("id"),data);
		INKFResponse resp=aContext.createResponseFrom(data);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);		
		
	}
	
	
	private void doSync(INKFRequestContext aContext) throws Exception
	{
		int availableProc=Runtime.getRuntime().availableProcessors();		
		int MAX_CONCURRENCY=32;
		String TEST="res:/sync/test2";
		//long PERIOD=500L;
		int N=4;
		
		int[] counts=new int[MAX_CONCURRENCY];
		long[] periods=new long[MAX_CONCURRENCY];
		int[] cpus=new int[MAX_CONCURRENCY];

		for (int n=0; n<N; n++)
		{
			for (int c=MAX_CONCURRENCY; c>=1; )
			{
				long period=1000+(MAX_CONCURRENCY-c+1)*1000/MAX_CONCURRENCY;

				WorkerThread[] threads=new WorkerThread[c];
				for (int t=0; t<c; t++)
				{	threads[t]=new WorkerThread(TEST,aContext);
				}
				for (int t=0; t<c; t++)
				{	threads[t].start();
				}
				Thread.sleep(200);
				sampleCPU();
				for (int t=0; t<c; t++)
				{	threads[t].begin();
				}
				Thread.sleep(period);
				cpus[c-1]+=sampleCPU();
				for (int t=0; t<c; t++)
				{	threads[t].finish();
				}
				int count=0;
				long duration=0;
				for (int t=0; t<c; t++)
				{	threads[t].join();
					count+=threads[t].getCount();
					duration+=threads[t].getDuration();
				}
				counts[c-1]+=count;
				periods[c-1]+=duration/c;
				System.out.println(c+" "+period+" "+duration/c);

				int k=2;
				if (k<1 || (c-k)<=availableProc) k=1;
				c-=k;
				
			}
		}
		
		HDSBuilder b=new HDSBuilder();
		b.pushNode("sync");
		for (int i=0; i<MAX_CONCURRENCY; i++)
		{	b.pushNode("test");
			int c=i+1;
			b.addNode("threadcount",c);
			long period=periods[i];
			if (period==0) period=1;
			long count=counts[i];
			long throughput=60000*count/period;
			long responseTime=(count>0)?period*c/count:0;
			b.addNode("response-time",responseTime);
			b.addNode("throughput",throughput);
			b.addNode("cpu",cpus[i]/N);
			
			b.popNode();
		}
		
		
		IHDSNode data=b.getRoot();
		aContext.sink("pds:/tools/nkmark10/test/"+aContext.getThisRequest().getArgumentValue("id"),data);
		INKFResponse resp=aContext.createResponseFrom(data);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	private static class WorkerThread extends Thread
	{
		private String mTest;
		private INKFRequestContext mContext;
		private boolean mFinish;
		private int mCount;
		private long mTime;
		public WorkerThread(String aTest, INKFRequestContext aContext)
		{	mTest=aTest;
			mContext=aContext;
		}
		public void begin()
		{	synchronized(this)
			{	notify();
			}
		}
		public void finish()
		{	mFinish=true;
		}
		public int getCount()
		{	return mCount;
		}
		public long getDuration()
		{	return mTime;
		}
		public void run()
		{	try
			{	INKFRequest req=mContext.createRequest(mTest);
				req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
				synchronized(this)
				{	wait();
				}
				mTime=System.currentTimeMillis();
				while(!mFinish)
				{	mContext.issueRequest(req);
					mCount++;
				}
			}
			catch (Exception e)
			{	System.out.println(e.toString());
			}
			mTime=System.currentTimeMillis()-mTime;
		}
		
	}
	
	private void doAsync(INKFRequestContext aContext) throws Exception
	{
		int originalThreadCount=aContext.getKernelContext().getKernel().getConfiguration().getInt("netkernel.scheduler.threadcount");

		int availableProc=Runtime.getRuntime().availableProcessors();
		int maxThreads=32; 
		long period=4000L;
		HDSBuilder b=new HDSBuilder();
		b.pushNode("async");
		
		
		int[] cpus = new int[maxThreads];
		long[] throughputs = new long[maxThreads];
		
		int N=3;
		for (int n=0; n<N; n++)
		{
			for (int t=maxThreads; t>=1; )
			{
				setKernelProperty("netkernel.scheduler.threadcount",Integer.toString(t),aContext);
				Thread.sleep(300);
				System.out.println("async("+t+")");
				
				sampleCPU();
				INKFRequest req=aContext.createRequest("active:timeTest");
				req.addArgumentByValue("identifier", "res:/async/test1");
				req.addArgumentByValue("period", period);
				req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
				
				long throughput=(Long)aContext.issueRequest(req);
				int cpu=sampleCPU();
				
				cpus[t-1]+=cpu;
				throughputs[t-1]+=throughput;
				
				int k=2;
				if (k<1 || (t-k)<=availableProc) k=1;
				t-=k;
			}
		}
		
		for (int i=0; i<maxThreads; i++)
		{
			b.pushNode("test");
			b.addNode("threadcount",i+1);
			long throughput=throughputs[i]/N;
			int cpu=cpus[i]/N;
			long responseTime=throughput>0?60000/throughput:0;
			b.addNode("throughput",throughput);
			b.addNode("cpu",cpu);
			b.addNode("response-time",responseTime);
			b.popNode();
		}
		
		b.popNode();
		
		setKernelProperty("netkernel.scheduler.threadcount",Integer.toString(originalThreadCount),aContext);
		IHDSNode data=b.getRoot();
		aContext.sink("pds:/tools/nkmark10/test/"+aContext.getThisRequest().getArgumentValue("id"),data);
		INKFResponse resp=aContext.createResponseFrom(data);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);

	}
	
	
	private void setKernelProperty(String aProperty,String aValue, INKFRequestContext aContext) throws Exception
	{	IKernel k=aContext.getKernelContext().getKernel();
		PropertyConfiguration pc=(PropertyConfiguration)k.getConfiguration();
		pc.setProperty(aProperty, aValue);
		((Kernel)k).setConfiguration(pc);
		
	}
	
	
	
	private long mLastNow;
	private long mLastGCTime;
	private static class ThreadTime
	{	public long mCPUTime;
	}	
	private Map<Long,ThreadTime> mThreadTimes=new HashMap<Long,ThreadTime>();
	
	private int sampleCPU()
	{
		long now=System.currentTimeMillis();
		long period=now-mLastNow;
		long cpus=Runtime.getRuntime().availableProcessors();
		float cpuTime=0;
		
		ThreadMXBean tmxb = ManagementFactory.getThreadMXBean( );
		if (tmxb.isThreadCpuTimeSupported() && tmxb.isThreadCpuTimeEnabled())
		{
			long[] threadIds=tmxb.getAllThreadIds();
			for (int i=0; i<threadIds.length; i++)
			{	long id=threadIds[i];
				long t2 = tmxb.getThreadCpuTime( id );
				Long lid=Long.valueOf(id);
				ThreadTime tt = mThreadTimes.get(lid);
				if (t2<0)
				{	if (tt!=null)
					{	mThreadTimes.remove(lid);
					}
				}
				else if (tt==null)
				{	tt= new ThreadTime();
					tt.mCPUTime=t2;
					mThreadTimes.put(lid, tt);
				}
				else
				{	cpuTime+=t2-tt.mCPUTime;
					tt.mCPUTime=t2;
				}
			}
			
			//gc time
			long gctime=0;
			List<GarbageCollectorMXBean> gcbeans=ManagementFactory.getGarbageCollectorMXBeans();
			for (GarbageCollectorMXBean gcbean: gcbeans)
			{	gctime+=gcbean.getCollectionTime();
			}
			long gcMs=gctime-mLastGCTime;
			mLastGCTime=gctime;		

			if (mLastNow>0)
			{						
				//System.out.println(gcMs);
				//System.out.println(cpuTime/1000000);
				cpuTime+=gcMs*1000000;
				cpuTime/=period*cpus;
				//cpuTime+=(gctime-mLastGCTime)*1000000;
				float cput=(float)cpuTime;
				cpuTime=cpuTime;
			}
		}
		mLastNow=now;
		
		
		
		return (int)(cpuTime/10000);
	}
	
	

}
