package org.netkernel.mod.performance.nkmark;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.ThreadMXBean;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.xml.xda.IXDAReadOnly;
import org.netkernel.xml.xda.IXDAReadOnlyIterator;
import org.netkernel.container.IKernel;
import org.netkernel.container.impl.Kernel;
import org.netkernel.layer0.nkf.INKFAsyncRequestHandle;
import org.netkernel.layer0.nkf.INKFExpiryFunction;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.PropertyConfiguration;
import org.netkernel.layer0.util.Utils;
import org.w3c.dom.Document;

public class NKMarkAccessor extends StandardAccessorImpl
{	
	private boolean mWarmedUp;	
	
	public NKMarkAccessor()
	{	declareThreadSafe();
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		String id=aContext.getThisRequest().getArgumentValue("id");
		if (id.equals(""))
		{	doTemplate(aContext);
		}
		else if (id.equals("chart"))
		{	doChart(aContext);
		}
		else if (id.equals("refresh"))
		{	aContext.delete("fpds:/tools/nkmark/test");
			doChart(aContext);
		}
		else if (id.equals("test"))
		{	if (!mWarmedUp)
			{	mWarmedUp=true;
				doMark(aContext);
			}
			doMark(aContext);
		}		
	}
	
	public void doTemplate(INKFRequestContext aContext) throws Exception
	{	INKFResponseReadOnly respIn=aContext.sourceForResponse("res:/org/netkernel/mod/performance/nkmark/template.xml");
		INKFResponse resp=aContext.createResponseFrom(respIn);
		resp.setHeader("WrappedControlPanel", true);		
	}

	public void doChart(INKFRequestContext aContext) throws Exception
	{	Document data=aContext.source("fpds:/tools/nkmark/test",Document.class);
		INKFRequest req=aContext.createRequest("active:xslt");
		req.addArgumentByValue("operand", data);
		req.addArgument("operator", "res:/org/netkernel/mod/performance/nkmark/style-mark.xsl");
		INKFResponseReadOnly respIn=aContext.issueRequestForResponse(req);
		aContext.createResponseFrom(respIn);
	}

	public void doWarmup(INKFRequestContext aContext) throws Exception
	{	mWarmedUp=true;
		//initialise/warmup
		mThreadTimes.clear();
		INKFRequest req=aContext.createRequest("active:timeTest");
		req.addArgumentByValue("identifier", "res:/async/test");
		req.addArgumentByValue("period", 5000L);
		aContext.issueRequest(req);		
	}
			
	private void doMark(INKFRequestContext aContext) throws Exception
	{
		int originalThreadCount=aContext.getKernelContext().getKernel().getConfiguration().getInt("netkernel.scheduler.threadcount");
		int availableProc=Runtime.getRuntime().availableProcessors();
		setKernelProperty("netkernel.scheduler.threadcount",Integer.toString(availableProc),aContext);	
		
		long period=5000L;
		HDSBuilder b=new HDSBuilder();
		b.pushNode("mark");
		
		
		//raw CPU processing
		/*
		{
			int c=availableProc;
			CPULoadThread[] threads=new CPULoadThread[c];
			for (int t=0; t<c; t++)
			{	threads[t]=new CPULoadThread();
			}
			for (int t=0; t<c; t++)
			{	threads[t].start();
			}
			Thread.sleep(200);
			sampleCPU();
			for (int t=0; t<c; t++)
			{	threads[t].begin();
			}
			Thread.sleep(5000);
			int cpu=sampleCPU();
			for (int t=0; t<c; t++)
			{	threads[t].finish();
			}
			for (int t=0; t<c; t++)
			{	threads[t].join();
			}
			long count=0;
			for (int t=0; t<c; t++)
			{	count+=threads[t].getCount();
			}
			b.addNode("raw-mark",count/64);
			b.addNode("raw-cpu",cpu);
			System.out.println("raw: "+count+" "+cpu);
		}
		*/
		
		/*
		//raw CPU 2
		{	INKFRequest req=aContext.createRequest("active:timeTest");
			req.addArgumentByValue("identifier", "active:cpuLoad");
			req.addArgumentByValue("period", period);
			req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
			long threadCount=availableProc;
			INKFAsyncRequestHandle[] handle = new INKFAsyncRequestHandle[(int)threadCount];
			sampleCPU();
			for (int t=0; t<threadCount; t++)
			{	handle[t]=aContext.issueAsyncRequest(req);
			}
			long throughput=0;
			for (int t=0; t<threadCount; t++)
			{	Long l=(Long)handle[t].join();
				long result=l;
				throughput+=result;
			}
			int cpu=sampleCPU();
			b.addNode("load-mark",throughput);
			b.addNode("load-cpu",cpu);
			System.out.println("load: "+throughput+" "+cpu);
		}
		*/
		
		
		
		
		System.gc();
		Thread.sleep(100);
		
		sampleCPU();
		
		//async performance
		long asyncMark;
		{	INKFRequest req=aContext.createRequest("active:timeTest");
			req.addArgumentByValue("identifier", "res:/async/test");
			req.addArgumentByValue("period", period);
			req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
			
			asyncMark=(Long)aContext.issueRequest(req);
			int cpu=sampleCPU();
			b.addNode("async-mark",asyncMark);
			b.addNode("async-cpu",cpu);
		}
		
		//synchronous performance
		System.gc();
		Thread.sleep(100);
		
		
		long syncMark;
		{	long threadCount=availableProc;
			INKFAsyncRequestHandle[] handle = new INKFAsyncRequestHandle[(int)threadCount];
			
			int N=4;
			int cpu=0;
			long maxThroughput=0;
			for (int i=0; i<N; i++)
			{
				sampleCPU();
				for (int t=0; t<threadCount; t++)
				{	
					INKFRequest req=aContext.createRequest("active:timeTest");
					req.addArgumentByValue("identifier", "res:/sync/test");
					req.addArgumentByValue("period", 1000L);
					req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
					handle[t]=aContext.issueAsyncRequest(req);
				}
				long throughput=0;
				for (int t=0; t<threadCount; t++)
				{	long result=(Long)handle[t].join();
					throughput+=result;
				}
				if (throughput>maxThroughput)
				{	maxThroughput=throughput;
					cpu=sampleCPU();
				}
			}
			
			syncMark=maxThroughput/6;
			b.addNode("sync-mark",syncMark);
			b.addNode("sync-cpu",cpu);
		}
		
		long gcMark;
		//GC performance
		{	long threadCount=availableProc*2;
			INKFAsyncRequestHandle[] handle = new INKFAsyncRequestHandle[(int)threadCount];
			MemoryPoolMXBean oldGen = Utils.getOldGen();
			long heapKb=oldGen.getUsage().getMax()/1024;
			long dataSize=heapKb/16000;
			//System.out.println("datasize="+dataSize);
			
			String id="res:/cache/CachableDataSet/"+dataSize+"/20000";

			for (int t=0; t<threadCount; t++)
			{	
				INKFRequest req=aContext.createRequest("active:timeTest");
				req.addArgumentByValue("identifier", id);
				req.addArgumentByValue("period", period);
				req.setHeader(INKFRequest.HEADER_FORGET_DEPENDENCIES, Boolean.TRUE);
				handle[t]=aContext.issueAsyncRequest(req);
			}
			long throughput=0;
			for (int t=0; t<threadCount; t++)
			{	long result=(Long)handle[t].join();
				throughput+=result;
			}
			int cpu=sampleCPU();
			
			gcMark=throughput/2000;
			b.addNode("gc-mark",gcMark);
			b.addNode("gc-cpu",cpu);
		}
		
		//Storage performance
		System.gc();
		Thread.sleep(100);
		sampleCPU();		
		
		long ioWriteMark;
		long ioReadMark;
		{
			int FILES=10;
			int SIZE=1024;
			int N=4;
			byte[] data=new byte[1024];
			File[] files=new File[FILES];
			for (int i=0; i<FILES; i++)
			{
				files[i]=File.createTempFile("nk_modperf", ".dat");
				files[i].deleteOnExit();
			}
			
			sampleCPU();
			long t0=System.currentTimeMillis();
			for (int j=0; j<N; j++)
			{	for (int i=0; i<FILES; i++)
				{
					FileOutputStream fos=new FileOutputStream(files[i]);
					BufferedOutputStream bos=new BufferedOutputStream(fos);
					for (int k=0; k<SIZE; k++)
					{	bos.write(data);
					}
					bos.flush();
					bos.close();
				}
			}
			long t1=System.currentTimeMillis();
			System.gc();
			Thread.sleep(100);
			int writeCPU=sampleCPU();
			long t2=System.currentTimeMillis();
			
			for (int j=0; j<N; j++)
			{	for (int i=0; i<FILES; i++)
				{
					FileInputStream fis=new FileInputStream(files[i]);
					BufferedInputStream bis=new BufferedInputStream(fis);
					int c;
					while ((c=bis.read(data))>0)
					{	
					}
					bis.close();
				}
			}
			long t3=System.currentTimeMillis();
			int readCPU=sampleCPU();
			int M=FILES*SIZE*N;
			ioWriteMark=M/(t1-t0);
			b.addNode("io-write-mark",ioWriteMark);
			b.addNode("io-write-cpu",writeCPU);
			ioReadMark=M/(t3-t2)/2;
			b.addNode("io-read-mark",ioReadMark);
			b.addNode("io-read-cpu",readCPU);
			
			for (int i=0; i<FILES; i++)
			{	files[i].delete();
			}
		}
		
		
		//restore kernel threadcount
		setKernelProperty("netkernel.scheduler.threadcount",Integer.toString(originalThreadCount),aContext);			
		
		b.addNode("time", new Date().toString());
		IHDSNode env= aContext.source("netkernel:/environment",IHDSNode.class);
		b.addNode("hostname", env.getFirstValue("/*/hostname"));
		long markLong=( (40*syncMark)+(20*asyncMark)+(15*gcMark)+(15*ioReadMark)+(10*ioWriteMark))/1000;
		float mark=((float)markLong)/10.0f;
		b.addNode("mark", mark);
		
		
		b.popNode();
		IHDSNode data=b.getRoot();
		aContext.sink("fpds:/tools/nkmark/test",data);
		aContext.createResponseFrom(data);
	}
		
	private void setKernelProperty(String aProperty,String aValue, INKFRequestContext aContext) throws Exception
	{	IKernel k=aContext.getKernelContext().getKernel();
		PropertyConfiguration pc=(PropertyConfiguration)k.getConfiguration();
		pc.setProperty(aProperty, aValue);
		((Kernel)k).setConfiguration(pc);
	}
	
	
	private long mLastNow;
	private long mLastGCTime;
	private static class ThreadTime
	{	public long mCPUTime;
	}	
	private Map<Long,ThreadTime> mThreadTimes=new HashMap<Long,ThreadTime>();
	
	private int sampleCPU()
	{
		long now=System.currentTimeMillis();
		long period=now-mLastNow;
		long cpus=Runtime.getRuntime().availableProcessors();
		float cpuTime=0;
		
		ThreadMXBean tmxb = ManagementFactory.getThreadMXBean( );
		if (tmxb.isThreadCpuTimeSupported() && tmxb.isThreadCpuTimeEnabled())
		{
			long[] threadIds=tmxb.getAllThreadIds();
			for (int i=0; i<threadIds.length; i++)
			{	long id=threadIds[i];
				long t2 = tmxb.getThreadCpuTime( id );
				Long lid=Long.valueOf(id);
				ThreadTime tt = mThreadTimes.get(lid);
				if (t2<0)
				{	if (tt!=null)
					{	mThreadTimes.remove(lid);
					}
				}
				else if (tt==null)
				{	tt= new ThreadTime();
					tt.mCPUTime=t2;
					mThreadTimes.put(lid, tt);
				}
				else
				{	cpuTime+=t2-tt.mCPUTime;
					tt.mCPUTime=t2;
				}
			}
			
			//gc time
			long gctime=0;
			List<GarbageCollectorMXBean> gcbeans=ManagementFactory.getGarbageCollectorMXBeans();
			for (GarbageCollectorMXBean gcbean: gcbeans)
			{	gctime+=gcbean.getCollectionTime();
			}
			long gcMs=gctime-mLastGCTime;
			mLastGCTime=gctime;		

			if (mLastNow>0)
			{						
				//System.out.println(gcMs);
				//System.out.println(cpuTime/1000000);
				cpuTime+=gcMs*1000000;
				cpuTime/=period*cpus;
				//cpuTime+=(gctime-mLastGCTime)*1000000;
				float cput=(float)cpuTime;
				cpuTime=cpuTime;
			}
		}
		mLastNow=now;
		
		
		
		return (int)(cpuTime/10000);
	}
	
	/*
	private static class CPULoadThread extends Thread
	{
		private volatile boolean mFinish;
		private long mCount;
		
		public void begin()
		{	synchronized(this)
			{	notify();
			}
		}
		public void finish()
		{	mFinish=true;
		}
		public long getCount()
		{	return mCount;
		}
		public void run()
		{	try
			{	synchronized(this)
				{	wait();
				}
				final int N=256;
				final int K=N*256;
				int[] b=new int[N];
				while(!mFinish)
				{	
					for (int i=0; i<K; i++)
					{	b[i%N]=i;
					}
					mCount++;
				}
			}
			catch (Exception e)
			{	System.out.println(e.toString());
			}
		}
		
	}
	*/

}
